------------------------------------------------
--Script: 		BouncyBones 
--Version:		1.00
--Created:      2012
--By:     		The Mind Forest
--Web:	      www.themindforest.com
--E-mail: 		info@themindforest.com

-- PLEASE give credit to Jonah Shafer if you use this script for any commercial purpose.
-- I have tried to add many comments to help make sense of what is going on.
------------------------------------------------

--// Initializing begins with setting the variable SpringBoneName to equal the text value "SpringDummy_"
SpringBoneName = "SpringDummy_"

-- Here, three arrays are created that I will store the main bouncy bones system objects in.
HelperArray = #()
BoneArray = #() 
ForcesArray = #()

-- What happens when we close a scene with bouncy bones after they have been created, or if we close the bouncy bones panel and open it again?  
-- The arrays are cleared and the bouncy bones systems elements cannot be edited except each part manually!  

-- If loading a scene with bouncy bones already applied, the bouncy bones systems objects need to be added back into the array...
-- We can see if a bouncy bones system was created previously in a scene by looking for specific bouncy bones named objects.  
-- This works because I am naming all the objects a custom name using the 'uniquename' code.
-- So..      If SpringDummy_001 exists  chances are bouncy bones was used and we should add all objects to array..  
-- This is the cheap and quick way of doing this and might be prone to causing errors potentially, but I am not sure of another way to go about re-registering the system.

-- This adds everything ( bones, dummies, forces ) back into their proper array if $SpringDummy 001 exists.
if ( $SpringDummy_001 != undefined )  then ( 
	
	select $SpringDummy_*
	join HelperArray $	
	
	select $BouncyBone_*
	join BoneArray $	
	
	select $SpringForce_*
	-- By adding the spring to the forces array container, I can select all the forces later for globally effecting all forces.
	join ForcesArray $
	deselect $
	) 
	else (
		messagebox "Welcome to bouncy bones!  Select two bones, root bone first, then dynamic bone second. Click Make Bouncy Bones button after setting parameters." title:"No Bouncy Bones in scene!" 
		)
		
--// Define the Function for 'affect pivot only' rotation
fn RotatePivotOnly obj rotation= ( local rotValInv=inverse (rotation as quat)
animate off in coordsys local obj.rotation*=RotValInv
obj.objectoffsetrot*=RotValInv
obj.objectoffsetpos*=RotValInv
)

-- Define a new Rollout..
Rollout ToolSection "[ Building Bouncy Bones ]"  
(

-- Now lets create all of the buttons in the first rollout, after the buttons are defined I can define what they do when interacted with.	
-- In some cases I am creating a label or name for a button or spinner and just giving it a slightly different name, strengthlabelx and  strengthlabel3 for example.  
--The unique name allows you to target it in the code.
label strengthlabelx ":: Configure Initial Spring Settings ::"align:#center
	label strengthlabel3 "============================="
spinner MassSets "Mass:" range:[1,10000,300] type:#integer fieldWidth: 40 across:2 align:#center
spinner DragSets "Drag:" range:[0,10,1] type:#float fieldWidth: 32 align:#center
--label strengthlabel0 "============================="
label strengthlabel1 " Spring Strength per Axis "
label strengthlabel2 "X       -        Y       -      Z"
spinner XStrongSets "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
spinner YStrongSets "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
spinner ZStrongSets "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
spinner VectSets "LookAt Vector Length:" range:[0,1000,0] type:#float fieldWidth: 45 align:#center
checkbox Hide_ck "Hide Helpers on Create?" checked:false
spinner SizeSets "Size:" range:[0,1000,0.3] type:#float fieldWidth: 35 across:2 align:#center
spinner OfSSets "Offset:" range:[0,1000,30] type:#integer fieldWidth: 35 align:#center
checkbox Wind_ck "Activate Dynamics on Create?" checked:true	
checkbox vertROT_ck "^  Maintain orientation?" checked:true
label strengthlabel4 "============================="
label axisflips	"Quick Rotate Object Pivot" 
button XFlipsButton "X" width:20 height: 20 across:4
button YFlipsButton "Y" width:20 height: 20
button ZFlipsButton "Z" width:20 height: 20
button HelpFlipsButton "?" width:20 height: 20
checkbox AutoFlip_ck "Auto Flip Axis on Create?" checked:false
button ApplyButton ">> Make Bouncy Bone <<" width:180 height: 50

-- With all of the buttons defined, I can now program what they do when interacted with.
on XFlipsButton pressed do ( if ( selection.count == 1 ) then ( 	
	toolMode.coordsys #local
	RotatePivotOnly $ (EulerAngles 0 -90 0)	))
on YFlipsButton pressed do ( if ( selection.count == 1 ) then ( 	
	toolMode.coordsys #local
	RotatePivotOnly $ (EulerAngles 0 0 -90) 	))
on ZFlipsButton pressed do ( if ( selection.count == 1 ) then ( 	
	toolMode.coordsys #local
	RotatePivotOnly $ (EulerAngles -90 0 0) 	))
	
-- This help button will open a little dialogue window.	
on HelpFlipsButton pressed do ( messagebox "Pressing these buttons automatically switches you to local coordinate system, so you can see the true direction of the objects pivot. To achieve proper results, it is required that you make sure the X axis of both objects pivots are pointing in the same direction from the parent object to the child." title:"Help"  )
	
	
-- What happens when the Make Bouncy Bones button is pressed?
on ApplyButton pressed do 
(
-- if there are two objects or bones selected, then bouncy bones are created, otherwise there is an error message.
if ( selection.count == 2 ) then ( 
	
-- For now this only works if two bones ( or objects ) are selected.
	
-- Switch from the default ( which is usually 'view') to 'local' coordinate translations
toolMode.coordsys #local

-- Create the array from the selection
BonesArray = getCurrentSelection()
	
-- Give the bones unique names..	
BonesArray[1].name = uniquename "BouncyBone_"
BonesArray[2].name = uniquename "BouncyBone_"

-- Make sure the bones are linked automatically with the first bone selected setup as the parent.
	-- This is especially useful for using this script on objects, which by default are not linked as bones are.
BonesArray[2].parent = BonesArray[1]

-- Add bones to the array.
join BoneArray $


-- There is a wierd bug with creating spring forces/controllers with maxscript, so for my solution a dummy object is loaded from a file which already has a spring applied and a wind force added in..
	getMAXFileObjectNames "BouncyBonesTool/SpringDummy.max" 
	mergeMaxFile "BouncyBonesTool/SpringDummy.max" --#select     --// selects imported
	
	--Dummy pos:[0,0,-10] isSelected:on  -- Before I was creating a dummy, and then trying to use <boolean>addSpring <node>node but I could not get it to work.
	
-- So I select the spring dummy, give it a new unique name, and then add it to the Helper Array..	
	select $SpringDummy
	$SpringDummy.name = uniquename SpringBoneName
	dumby = getCurrentSelection()	
	join HelperArray dumby[1]
	attachObjects BonesArray[1] dumby[1]
	
	if hide_ck.checked == true do 
	( hide $ ) --hide $  -- Option to hide the spring DUMMIES.
	
	--dumby[1].position.controller = SpringPositionController ()  -- here is where the spring position controller was being added from before, however, since it cannot be added this way without issues, it is applied by loading in a dummy with it already applied and saved in a file.

	-- because the spring position controller was already created, we can affect the controllers values
 	dumby[1].position.controller.setMass MassSets.value
 	dumby[1].position.controller.setDrag DragSets.value
	dumby[1].position.controller.x_effect = XStrongSets.value
	dumby[1].position.controller.y_effect = YStrongSets.value
	dumby[1].position.controller.z_effect = ZStrongSets.value
	dumby[1].position.controller.steps = 2
 	dumby[1].pos = BonesArray[1].pos
 	dumby[1].rotation.x_rotation = BonesArray[1].rotation.x_rotation
 	dumby[1].rotation.y_rotation = BonesArray[1].rotation.y_rotation
 	dumby[1].rotation.z_rotation = BonesArray[1].rotation.z_rotation
	
	if Wind_ck.checked == false do
	(dumby[1].position.controller.steps = 0)


in coordsys local move dumby[1] [OfSSets.value,0,0]

-- Create a new dummy which is only used for setup, later deleted..
Dummy pos:[0,0,-10] isSelected:on 
	dumby2 = getCurrentSelection() 
	dumby2[1].position = BonesArray[2].position 
	dumby2[1].rotation.x_rotation = BonesArray[2].rotation.x_rotation
	dumby2[1].rotation.y_rotation = BonesArray[2].rotation.y_rotation
	dumby2[1].rotation.z_rotation = BonesArray[2].rotation.z_rotation

in coordsys local move dumby2[1] [OfSSets.value,0,0]	
    BonesArray[2].rotation.controller = LookAt_Constraint ()
	BonesArray[2].rotation.controller.appendTarget dumby[1] 100
    BonesArray[2].rotation.controller.upnode_world = off
	BonesArray[2].rotation.controller.pickUpNode  = dumby[1] 



---/////////////////////////////////  Rotation methods based on the Maintain Orientation check box... 	
	if vertROT_ck.checked == false do
			(
			--print "Applying Branch Rotation"
			BonesArray[2].rotation.controller.target_axis = 0  -- X
			BonesArray[2].rotation.controller.StoUP_axis = 0  -- Y W/Flip defined below in If
			BonesArray[2].rotation.controller.upnode_axis = 0  -- Z
			BonesArray[2].rotation.controller.upnode_ctrl = 1 
			
			if AutoFlip_ck.checked == true then	(BonesArray[2].rotation.controller.StoUP_axisFlip = true ) 
			else (BonesArray[2].rotation.controller.StoUP_axisFlip = false)
			)
			
	if vertROT_ck.checked == true do
			(
			--print "Applying Trunk Rotation"
			BonesArray[2].rotation.controller.target_axis = 0  -- 0=X 1=Y 2=Z  -- Select LookAt Axis
			BonesArray[2].rotation.controller.StoUP_axis = 2   -- 0=X 1=Y 2=Z 
			BonesArray[2].rotation.controller.upnode_axis = 2  -- 0=X 1=Y 2=Z
			BonesArray[2].rotation.controller.upnode_ctrl = 1
			BonesArray[2].rotation.controller.StoUP_axisFlip = off
				
			if AutoFlip_ck.checked == true then	(BonesArray[2].rotation.controller.StoUP_axisFlip = true ) 
			else (BonesArray[2].rotation.controller.StoUP_axisFlip = false)
			)
			
---/////////////////////////////////
	BonesArray[2].rotation.controller.viewline_length_abs = true
	BonesArray[2].rotation.controller.lookat_vector_length = VectSets.value
	dumby[1].pos = dumby2[1].pos
	dumby[1].rotation.x_rotation = dumby2[1].rotation.x_rotation
	dumby[1].rotation.y_rotation = dumby2[1].rotation.y_rotation
	dumby[1].rotation.z_rotation = dumby2[1].rotation.z_rotation
	dumby[1].scale = [SizeSets.value,SizeSets.value,SizeSets.value]
	
			
	select dumby[1]
	--max tool zoomextents all
			
	-- Select the spring force ( wind object) that is loaded into the file along with the spring dummy.
	select $springforce
   -- reposition the force object to the second dummy position ( otherwise all the wind forces are positioned at world space 0,0,0 on top of each other)	
	$springforce.pos = dumby2[1].pos
   attachObjects BonesArray[1] $springforce 
	
	$springforce.frequency = 0.001
	--Rename all spring forces to a new unique name, this allows me to load in the original springforce from file again and again.
	$springforce.name = uniquename "SpringForce_"
	-- By adding the spring to the forces array container, I can select all the forces later for globally effecting all forces.
	join ForcesArray $
	
	hide $
			
			-- Ther e is a second dummy made during the setup process, which is now deleted..
			delete dumby2
			
	) 
	else (messagebox "Must select TWO linked bones or Objects. The object selected first will be the root, and the second object selected will have bouncy dynamics applied." title:"Error")
	-- END selection.count IF
)
-- END APPLY BUTTON


)--// Close Rollout

-- Define a new Rollout..  Rolled up is true so it is hidden at the start.
Rollout WindDynSection "[ Post-Build Control ]" rolledup: true
(
	
-- Select all helpers by selecting the array.	
button SelectButton "Select All Helpers" width:180 align:#center
on SelectButton pressed do (select HelperArray)

button NoiseRootButton "Apply Root Noise" width:180 align:#center
on NoiseRootButton pressed do 
(

	if ( selection.count == 1) then (
	

    lastselected = getCurrentSelection()
	lastselectedpos =	$.pos
	--print	lastselectedpos
		
	Dummy pos:[0,0,0] isSelected:on 
		
	$.name = "NoiseDummyMammy" -- The mother of all noise, lol.
	--$.pos = lastselected.pos
	$.rotation.controller = noise_rotation ()
	$.rotation.controller.noise_strength.x = 0.5
    $.rotation.controller.noise_strength.y = 0.5
    $.rotation.controller.noise_strength.z = 0.5
    $.rotation.controller.frequency = 0.003
	
	select lastselected
	$.rotation.controller = Orientation_Constraint()
    $.rotation.controller.appendTarget $NoiseDummyMammy 100
    $.rotation.controller.relative = on
		
	select $NoiseDummyMammy
	$.pos = lastselectedpos
	$.name = uniquename "NoiseDummy_"
	
	--hide $ -- I was going to hide the noise, but this is turned off because clicking on the dummy is the only way to change the settings of the random noise controller for now after it is made.
	
) 
-- This is what happens if you have more or less than two objects selected when you press the Make Bouncy Bones button.
else messagebox "Select a single bone or object first, then a dummy with a noise controller will be made and linked to the selected object." title:"Unsuccessfully noisy!"
	
	
) 




label helplabel0 "============================="
label helplabel1 "- Automatic Global Wind Control -"
label helplabel2 "( for subtle random movement )"
spinner TurbSets "Wind Turbulence:" range:[0,9999999,0.05] type:#float fieldWidth: 70 align:#right
spinner FreqSets "Wind Frequency:" range:[0,9999999,0.001] type:#float fieldWidth: 70 align:#right
spinner SclSets "Wind Scale:" range:[0,9999999,2.0] type:#float fieldWidth: 70 align:#right
label RTlabel0 "============================="
label RTlabel1 "Global Dynamic Spring Strengths"
spinner MassSetsx "Mass:" range:[1,10000,300] type:#integer fieldWidth: 40 across:2 align:#center
spinner DragSetsx "Drag:" range:[0,10,1] type:#float fieldWidth: 32 align:#center
label RTlabel2 "Overides individual axis settings"
label RTlabel3 "X       -        Y       -      Z"
spinner XStrongSetsRT "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
spinner YStrongSetsRT "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
spinner ZStrongSetsRT "" range:[0,1000,50] type:#float fieldWidth: 45 across:3 align:#center
label RTlabel4 "============================="
label RTlabel4a "- Must have a helper node selected -"


on TurbSets changed val do ( for element in ForcesArray do element.turbulence = val  )
on FreqSets changed val do ( for element in ForcesArray do element.frequency = val  )
on SclSets changed val do ( for element in ForcesArray do element.baseobject.scale = val  )  -- there is a conflict with using just .scale because it is trying to affect the scale of the object not the scale of the effect of the wind.  The scale of the object is point3, which is X Y AND Z values..

--  OK.  So for some random reason, this code and the three lines below have to have a different ending...  In the top two lines, the = sign must not be included?
on MassSetsx changed val do ( for element in HelperArray do element.position.controller.setMass MassSetsx.value   )
on DragSetsx changed val do ( for element in HelperArray do element.position.controller.setDrag DragSetsx.value  )

on XStrongSetsRT changed val do ( for element in HelperArray do element.position.controller.x_effect = XStrongSetsRT.value  )
on YStrongSetsRT changed val do ( for element in HelperArray do element.position.controller.y_effect = YStrongSetsRT.value  )
on ZStrongSetsRT changed val do ( for element in HelperArray do element.position.controller.z_effect = ZStrongSetsRT.value  )
---////////////////////////


button allActiveWindButton "Activate All Dynamics" width:180 align:#center
on allActiveWindButton pressed do ( for element in HelperArray do element.position.controller.steps = 2 )

button allDActiveWindButton "De-activate All Dynamics" width:180 align:#center
on allDActiveWindButton pressed do ( for element in HelperArray do element.position.controller.steps = 0 )

button ActiveWindButton "Activate Individual Dyn." width:180 align:#center
on ActiveWindButton pressed do ( 

		if (selection.count == 1 ) then ( 
		if ( matchPattern $.name pattern:"SpringDummy_*" == true ) then ( $.position.controller.steps = 2 ) else (messagebox "Must Select a SpringDummy_### Helper." title:"Error")
 		--print "YEAYA"
 		) 
		else (messagebox "Must Select a Single SpringDummy_### Helper." title:"Error")
	)

button DActiveWindButton "De-activate Individual Dyn." width:180 align:#center
on DActiveWindButton pressed do ( 

		if (selection.count == 1 ) then ( 
		if ( matchPattern $.name pattern:"SpringDummy_*" == true ) then ( $.position.controller.steps = 0  )	else (messagebox "Must Select a SpringDummy_### Helper." title:"Error")
 		--print "YEAYA"
 		) 
		else (messagebox "Must Select a Single SpringDummy_### Helper." title:"Error")
	)
label RTlabel5 "============================="
label RTlabel6 "- Generated Elements Visibility -"	
button ShowWindButton "Show Wind" width:80 align:#center across:2
on ShowWindButton pressed do ( 
	if (ForcesArray != undefined) then ( unhide ForcesArray  ) else(messagebox "Bouncy Bone wind spacewarp doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper wind object." title:"Error")
	) 

button HideWindButton "Hide Wind" width:80 align:#center
on HideWindButton pressed do ( 
	if (ForcesArray != undefined) then ( hide ForcesArray  ) else(messagebox "Bouncy Bone wind spacewarp doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper wind object." title:"Error")
	) 
 	
button ShowHlprButton "Show Helper" width:80 align:#center across:2
on ShowHlprButton pressed do ( 
	if (HelperArray != undefined) then ( unhide HelperArray  ) else(messagebox "Bouncy Bone Helper Dummy doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper dummy object." title:"Error")
	) 

button HideHlprButton "Hide Helper" width:80 align:#center
on HideHlprButton pressed do ( 
	if (HelperArray != undefined) then ( hide HelperArray  ) else(messagebox "Bouncy Bone Helper Dummy doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper dummy object." title:"Error")
	) 
	
button ShowBoneButton "Show Bones" width:80 align:#center across:2
on ShowBoneButton pressed do ( 
	if (BoneArray != undefined) then ( unhide BoneArray  ) else(messagebox "Bouncy bone doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper bone objects." title:"Error")
	) 

button HideBoneButton "Hide Bones" width:80 align:#center
on HideBoneButton pressed do ( 
	if (BoneArray != undefined) then ( hide BoneArray  ) else(messagebox "Bouncy bone doesn't exist yet! Apply Bouncy Bones to two linked bones to automatically create the proper bone objects." title:"Error")
	) 
	
	--BoneArray
label next1 "============================="	
label next2 "- Flip selected bouncy bone axis -" 
	
button FixButton3 "Flip Rotation Axis" width:180 align:#center
on FixButton3 pressed do (

-- Need to find a way to check not only name but also if the $ has a rotation controller --  maybe get rotation controller?
	if ( matchPattern $.name pattern:"BouncyBone_*" == true ) then ( 

		if ($.rotation.controller.StoUP_axisFlip == true) 
		then ($.rotation.controller.StoUP_axisFlip = false)
		else ( $.rotation.controller.StoUP_axisFlip = true )

		)
	else (messagebox "Must have bouncy bones selected." title:"Error")

)
	
-- These buttons allow you to flip the way the look at constraints axis is oriented for the bouncy bone.
button FixButtona "Rotation on X" width:180 align:#center
on FixButtona pressed do (
			--print "Applying Branch Rotation"
			$.rotation.controller.StoUP_axis = 0   -- 0=X 1=Y 2=Z 
			$.rotation.controller.upnode_axis = 0  -- 0=X 1=Y 2=Z
			$.rotation.controller.upnode_ctrl = 1
			)

button FixButtonb "Rotation on Y" width:180 align:#center
on FixButtonb pressed do (
			--print "Applying Branch Rotation"
			$.rotation.controller.StoUP_axis = 1   -- 0=X 1=Y 2=Z 
			$.rotation.controller.upnode_axis = 1  -- 0=X 1=Y 2=Z  
			$.rotation.controller.upnode_ctrl = 1
			)

button FixButtonc "Rotation on Z" width:180 align:#center
on FixButtonc pressed do (
			--print "Applying Branch Rotation"

			$.rotation.controller.StoUP_axis = 2   -- 0=X 1=Y 2=Z 
			$.rotation.controller.upnode_axis = 2  -- 0=X 1=Y 2=Z  	  
			$.rotation.controller.upnode_ctrl = 1
			)		
			
			
			
button BoneButton "Open Edit Bones Panel" width:180 align:#center
on BoneButton pressed do (	macros.run "Animation Tools" "BoneAdjustmentTools"	)

	
)-- // END ROLLOUT

-- Define a new Rollout..  Rolled up is true so it is hidden at the start.
Rollout GuideSection "[ Quick Guide ]" rolledup: true
(
label guidelabelx "Bouncy Bones Application Method:"
label guidelabel0 "---" align:#center
label guidelabel1 "Two linked bones, no more or less."
label guidelabel2 "Select the parent root bone first."
label guidelabel3 "Hold ctrl then select dynamic bone."
label guidelabel4 "Choose settings, then press Apply."
label guidelabel11 "For objects instead of bones, you"
label guidelabel12 "must rotate the pivot axis so that"
label guidelabel13 "the local X points in the direction"
label guidelabel14 "from the parent to the child."
label guidelabel5 "---" align:#center
label guidelabel6 "More mass enhances gravity effect."
label guidelabel7 "More drag smoothes effect motion."
label guidelabel8 "Lower drag makes it more bouncy."
label guidelabel9 "---" align:#center
	
label guidelabel15 "Apply post build root noise after "
label guidelabel16 "applying with two bones selected."
label guidelabel10 "Read BouncyBonesUsageNotes.txt for more info."


)-- // END ROLLOUT

-- Define a new Rollout..  Rolled up is true so it is hidden at the start.
Rollout AboutSection "[ About Info ]" rolledup: true
(  
	bitmap teh_bmp fileName:"BouncyBones.bmp" 
    label lab1 "Bouncy Bones v1.00"
	hyperlink lab2_ab "Created by The Mind Forest" align:#center address:" mailto:info@themindforest.com" color:(color 0 0 0) hoverColor:(color 255 255 255)
	label lab3_ab "Copyright \xa9 2012" 
	hyperlink lab4_ab "View Website" align:#center address:"http://www.themindforest.com" color:(color 70 70 255) hoverColor:(color 255 255 255)
)-- // END ROLLOUT


-- This code is what actually makes the floater and window.
BOUNCYBONESTOOLS_floater = newRolloutFloater "Bouncy Bones Toolkit" 200 470
BOUNCYBONESTOOLS_floater.pos = [2,86]
addrollout ToolSection BOUNCYBONESTOOLS_floater
addrollout WindDynSection BOUNCYBONESTOOLS_floater
addrollout GuideSection BOUNCYBONESTOOLS_floater
addrollout AboutSection BOUNCYBONESTOOLS_floater